# -*- coding: utf-8 -*-
# (c) 2018-2020 Andreas Motl <andreas@hiveeyes.org>
# License: GNU Affero General Public License, Version 3
import json
import logging
from docopt import docopt, DocoptExit

from apicast import __appname__, __version__
from apicast.util import normalize_options, setup_logging
from apicast.core import dwd_beeflight_forecast_data, dwd_beeflight_forecast_stations, dwd_beeflight_site_url_by_slug, \
    dwd_beeflight_forecast_stations_site_slugs
from apicast.format import Formatter

log = logging.getLogger(__name__)


def run():
    """
    Access bee flight forecast information published by Deutscher Wetterdienst (DWD).

    Usage:
      apicast beeflight stations [--site-slugs]
      apicast beeflight forecast --url=<url> [--format=<format>]
      apicast beeflight forecast --station=<station> [--format=<format>]
      apicast service [--listen=<listen>]
      apicast --version
      apicast (-h | --help)

    Options:
      --url=<url>                       URL to detail page
      --station=<station>               Station identifier
      --format=<format>                 Output format: "json" or "table". Default: json
      --listen=<listen>                 HTTP server listen address. [Default: localhost:24640]
      --version                         Show version information
      --debug                           Enable debug messages
      -h --help                         Show this screen

    Examples::

        # Display list of stations
        apicast stations

        # Display bee flight forecast for Potsdam
        apicast --station=berlin_brandenburg/potsdam

    """

    name = f'{__appname__} {__version__}'

    # Parse command line arguments
    options = normalize_options(docopt(run.__doc__, version=name))

    # Setup logging
    debug = options.get('debug')
    log_level = logging.INFO
    if debug:
        log_level = logging.DEBUG
    setup_logging(log_level)

    # Debugging
    log.debug('Options: {}'.format(json.dumps(options, indent=4)))

    # Run service.
    if options.service:
        listen_address = options.listen
        log.info(f'Starting {name}')
        log.info(f'Starting web service on {listen_address}')
        from apicast.api import start_service
        start_service(listen_address)
        return

    # Run command.
    if options.stations:

        if options.site_slugs:
            result = dwd_beeflight_forecast_stations_site_slugs()

        else:
            result = dwd_beeflight_forecast_stations()

        print(json.dumps(result, indent=4))

    # Fetch and extract forecast information.
    elif options.url:
        result = dwd_beeflight_forecast_data(options.url)
        format_beeflight_forecast(result, options.format)

    elif options.station:
        url = dwd_beeflight_site_url_by_slug(options.station)
        result = dwd_beeflight_forecast_data(url)
        format_beeflight_forecast(result, options.format)


def format_beeflight_forecast(result, format='json'):

    data = result['data']
    if not data:
        raise ValueError('No data found or unable to parse')

    format = format or 'json'
    format = format.lower()

    if format == 'json':
        response = Formatter(result).normalize()
        print(json.dumps(response, indent=4))

    elif format == 'json-machine':
        response = Formatter(result).machinify()
        print(json.dumps(response, indent=4))

    elif format == 'table-markdown':
        response = Formatter(result).table_markdown()
        print(response)

    else:
        raise ValueError('Unknown output format. Please specify "json", "json-machine" or "table-markdown".')



# -*- coding: utf-8 -*-
# (c) 2020 Andreas Motl <andreas@hiveeyes.org>
import io
from contextlib import redirect_stdout

import tabulate
from datetime import datetime
from dateutil.parser import parse as parsedate


class Formatter:
    """
    Die Bedingungen für den Bienenflug werden in 5 Intensitätsstufen angegeben
    (kein - gering - mittel - stark - intensiv).

    -- https://www.dwd.de/DE/fachnutzer/freizeitgaertner/dokumentation/gw_bienenflug
    """

    LABEL_MAP = {
        'Datum': 'date',

        'morgens': 'morning',
        'mittags': 'noon',
        'abends': 'evening',

        'kein': 'no',
        'gering': 'low',
        'mittel': 'medium',
        'stark': 'strong',
        'intensiv': 'intensive',
    }

    STRENGTH_MACHINE_MAP = {
        'no': 0,
        'low': 1,
        'medium': 2,
        'strong': 3,
        'intensive': 4,
    }

    def __init__(self, result):
        self.result = result
        self.data = result["data"]

    def translate(self):
        for item in self.data:
            for index, slot in enumerate(item):
                for key, value in self.LABEL_MAP.items():
                    slot = slot.replace(key, value)
                item[index] = slot
        return self

    def normalize(self):
        result = []
        for item in self.data[1:]:
            item = dict(zip(self.data[0], item))
            result.append(item)
        return result

    def machinify(self):
        self.translate()
        data = self.normalize()
        for item in data:
            for key, value in item.items():
                if key == 'date':
                    value += str(datetime.now().year)
                    value = parsedate(value, fuzzy=True).strftime('%Y-%m-%d')
                elif value in self.STRENGTH_MACHINE_MAP.keys():
                    value = self.STRENGTH_MACHINE_MAP[value]
                item[key] = value
        return data

    def table_markdown(self):
        with io.StringIO() as buffer, redirect_stdout(buffer):

            # Report about weather station / observation location
            print(u'### Prognose des Bienenfluges in {}'.format(self.result['station']))
            print()

            # Output forecast data
            print(tabulate.tabulate(self.data[1:], headers=self.data[0], showindex=False, tablefmt='pipe'))

            return buffer.getvalue()



# -*- coding: utf-8 -*-
# (c) 2020 Andreas Motl <andreas@terkin.org>
# License: GNU Affero General Public License, Version 3
import logging
from copy import deepcopy

from fastapi import FastAPI, Query
from fastapi.responses import HTMLResponse, PlainTextResponse

from apicast import __appname__, __version__
from apicast.core import dwd_beeflight_forecast_stations, dwd_beeflight_forecast_stations_site_slugs, \
    dwd_beeflight_site_url_by_slug, dwd_beeflight_forecast_data, dwd_source, dwd_copyright, producer_name, producer_link
from apicast.format import Formatter

app = FastAPI()

log = logging.getLogger(__name__)


@app.get("/", response_class=HTMLResponse)
def index():
    appname = f'{__appname__} {__version__}'
    about = 'Apicast acquires bee flight forecast information published by Deutscher Wetterdienst (DWD).'
    return f"""
    <html>
        <head>
            <title>{appname}</title>
        </head>
        <body>
            <h3>About</h3>
            {about}
            <ul>
            <li>Source: DWD » Freizeitgärtner » Gartenwetter - <a href="{dwd_source}">{dwd_source}</a></li>
            <li>Producer: {producer_name} - <a href="{producer_link}">{producer_link}</a></li>
            <li>Data copyright: {dwd_copyright}</li>
            </ul>
            <h3>Index</h3>
            <ul>
            <li><a href="beeflight/stations/germany">List of federal states / sites</a></li>
            <li><a href="beeflight/stations/germany/site-slugs">List of site slugs</a></li>
            </ul>
            <h3>Examples</h3>
            <ul>
            <li><a href="beeflight/forecast/germany/berlin_brandenburg/potsdam">Bee flight forecast for "berlin_brandenburg/potsdam"</a></li>
            <li><a href="beeflight/forecast/germany/bayern/regensburg">Bee flight forecast for "bayern/regensburg"</a></li>
            </ul>
        </body>
    </html>
    """


@app.get("/robots.txt", response_class=PlainTextResponse)
def robots():
    return f"""
User-agent: *
Disallow: /beeflight/
    """.strip()


@app.get("/beeflight/stations/germany")
def beeflight_stations():
    stations = dwd_beeflight_forecast_stations()
    return make_json_response(stations)


@app.get("/beeflight/stations/germany/site-slugs")
def beeflight_stations_site_slugs():
    slugs = dwd_beeflight_forecast_stations_site_slugs()
    return make_json_response(slugs)


@app.get("/beeflight/forecast/germany/{state}/{site}")
def beeflight_forecast_by_slug(state: str, site: str, format: str = Query(default='json'), translate: bool = Query(default=False)):
    station_slug = f"{state}/{site}"

    try:
        url = dwd_beeflight_site_url_by_slug(station_slug)
        result = deepcopy(dwd_beeflight_forecast_data(url))
        data = result['data']
        if not data:
            raise ValueError('No data found or unable to parse')
    except Exception as ex:
        return {'error': str(ex)}

    formatter = Formatter(result)

    if translate:
        formatter.translate()

    if format == 'json-machine':
        response = formatter.machinify()

    elif format == 'table-markdown':
        return PlainTextResponse(formatter.table_markdown())

    else:
        response = formatter.normalize()

    return make_json_response(response)


def make_json_response(data):
    response = {
        'meta': {
            'source': dwd_source,
            'copyright': dwd_copyright,
            'producer': f'{producer_name} - {producer_link}',
        },
        'data': data,
    }
    return response


def start_service(listen_address):
    host, port = listen_address.split(':')
    port = int(port)
    from uvicorn.main import run
    run(app='apicast.api:app', host=host, port=port, reload=True)



# -*- coding: utf-8 -*-
# (c) 2018-2020 Andreas Motl <andreas@hiveeyes.org>
"""
Apicast acquires bee flight forecast information published by Deutscher Wetterdienst (DWD).

Parse information about "Gartenwetter » Bienenflug"
from https://www.dwd.de/DE/fachnutzer/freizeitgaertner/1_gartenwetter/_node.html ff.

See also https://community.hiveeyes.org/t/dwd-prognose-bienenflug/787


Synopsis::

    apicast --url https://www.dwd.de/DE/fachnutzer/freizeitgaertner/1_gartenwetter/berlin_brandenburg/potsdam/_node.html

"""
import re
import ttl_cache
import mechanicalsoup
from munch import Munch
from html_table_extractor.extractor import Extractor


user_agent = "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_7_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/49.0.2623.112 Safari/537.36"
dwd_source = "https://www.dwd.de/DE/fachnutzer/freizeitgaertner/1_gartenwetter/_node.html"
dwd_copyright = "© Deutscher Wetterdienst (DWD), Abteilung Agrarmeteorologie"
producer_name = "Apicast"
producer_link = "https://github.com/hiveeyes/apicast"


@ttl_cache(60 * 60 * 24)
def dwd_beeflight_forecast_stations():

    base_url = "https://www.dwd.de/"
    app_path = "DE/fachnutzer/freizeitgaertner/1_gartenwetter/"
    index_url = f"{base_url}{app_path}_node.html"

    browser = mechanicalsoup.StatefulBrowser(user_agent=user_agent)
    page = browser.open(index_url)

    # Find navigation section and extract elements.
    nav_main = page.soup.find(attrs={'class': 'navMain'})
    state_links = nav_main.find_all('a', href=re.compile('.+1_gartenwetter.+'))

    results = []

    for state_link in state_links:
        state_title = state_link['title']
        state_url = f"{base_url}{state_link['href']}"

        state_link_prefix = re.sub('/_node.html.*', '', state_link['href'])
        state_slug = state_link_prefix.replace(app_path, '')

        state = Munch({
            'title': state_title,
            'url': re.sub(';jsessionid=.*', '', state_url),
            'slug': state_slug,
            'sites': [],
        })

        page = browser.open(state_url)

        nav_main = page.soup.find(attrs={'class': 'navMain'})
        site_links = nav_main.find_all('a', href=re.compile(f'.*{state_link_prefix}.*'))

        for site_link in site_links:
            site_title = site_link['title']
            site_url = f"{base_url}{site_link['href']}"
            site_link_prefix = re.sub('/_node.html.*', '', site_link['href'])
            site_slug = site_link_prefix.replace(app_path, '')
            site = Munch({
                'title': site_title,
                'url': site_url,
                'slug': site_slug,
            })

            state.sites.append(site)

        results.append(state)

    return results


def dwd_beeflight_forecast_stations_site_slugs():
    stations = dwd_beeflight_forecast_stations()
    slugs = []
    for station in stations:
        for site in station.sites:
            slugs.append(site.slug)
    return slugs


def dwd_beeflight_site_url_by_slug(slug):
    url = f"https://www.dwd.de/DE/fachnutzer/freizeitgaertner/1_gartenwetter/{slug}/_node.html"
    return url


@ttl_cache(60)
def dwd_beeflight_forecast_data(url):

    # Navigate to HTTP resource.
    browser = mechanicalsoup.StatefulBrowser(user_agent=user_agent)
    page = browser.open(url)

    # Find content section and extract elements.
    subject = page.soup.find(string=u'Prognose des Bienenfluges')
    if not subject:
        raise ValueError('No forecast available for this station')
    station = subject.find_next('br').next.strip()
    table = subject.parent.find_next_sibling('table')
    # TODO: Read table footer "© Deutscher Wetterdienst, erstellt 12.04.2018 04:00 UTC"

    # Read HTML table
    data = parse_html_table(table)

    # Ready.
    result = {
        'station': station,
        'data': data,
    }
    return result


def parse_html_table(html):
    extractor = Extractor(html)
    extractor.parse()
    return extractor.return_list()



# -*- coding: utf-8 -*-
import os
from io import open
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, "README.rst"), encoding="UTF-8").read()

setup(
    name="apicast",
    version="0.7.0",
    description="Python client to access bee flight forecast information published by Deutscher Wetterdienst (DWD), "
    "the federal meteorological service in Germany.",
    long_description=README,
    license="AGPL 3, EUPL 1.2",
    classifiers=[
        "Programming Language :: Python",
        "License :: OSI Approved :: European Union Public Licence 1.2 (EUPL 1.2)",
        "License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)",
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Science/Research",
        "Topic :: Communications",
        "Topic :: Database",
        "Topic :: Internet",
        "Topic :: Scientific/Engineering :: Atmospheric Science",
        "Topic :: Scientific/Engineering :: GIS",
        "Topic :: Scientific/Engineering :: Human Machine Interfaces",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Interface Engine/Protocol Translator",
        "Topic :: Scientific/Engineering :: Visualization",
        "Topic :: Software Development :: Libraries",
        "Topic :: System :: Archiving",
        "Topic :: Text Processing",
        "Topic :: Utilities",
        "Operating System :: POSIX",
        "Operating System :: Unix",
        "Operating System :: MacOS",
    ],
    author="Andreas Motl",
    author_email="andreas@hiveeyes.org",
    url="https://github.com/hiveeyes/apicast",
    keywords="bee flight forecast information "
    "dwd cdc deutscher wetterdienst climate data center weather "
    "opendata data acquisition transformation export "
    "geospatial temporal timeseries "
    "sensor network observation "
    "http rest api "
    "json csv",
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        "MechanicalSoup==0.12.0",
        "html-table-extractor==1.4.1",
        "tabulate==0.8.7",
        "docopt==0.6.2",
        "munch==2.5.0",
        "ttl-cache==1.6",
        "python-dateutil>=2.8.0",
    ],
    extras_require={
        'service': [
            'fastapi==0.55.1',
            'uvicorn==0.11.5',
        ],
    },
    entry_points={"console_scripts": ["apicast = apicast.cli:run"]},
)
